@@ -1,4 +1,4 @@
-/* See LICENSE file for license and copyright information */
+/* SPDX-License-Identifier: Zlib */
 
 #define _POSIX_SOURCE
 #define _XOPEN_SOURCE 500
@@ -24,9 +24,10 @@
 
 #define KEY_PAGE                  "page"
 #define KEY_OFFSET                "offset"
-#define KEY_SCALE                 "scale"
+#define KEY_ZOOM                  "zoom"
 #define KEY_ROTATE                "rotate"
 #define KEY_PAGES_PER_ROW         "pages-per-row"
+#define KEY_PAGE_RIGHT_TO_LEFT    "page-right-to-left"
 #define KEY_FIRST_PAGE_COLUMN     "first-page-column"
 #define KEY_POSITION_X            "position-x"
 #define KEY_POSITION_Y            "position-y"
@@ -35,21 +36,46 @@
 
 #ifdef __GNU__
 #include <sys/file.h>
-#define file_lock_set(fd, cmd) flock(fd, cmd)
+
+#define FILE_LOCK_WRITE LOCK_EX
+#define FILE_LOCK_READ LOCK_SH
+
+static int
+file_lock_set(int fd, int cmd)
+{
+  return flock(fd, cmd);
+}
 #else
-#define file_lock_set(fd, cmd) \
-  { \
-  struct flock lock = { .l_type = cmd, .l_start = 0, .l_whence = SEEK_SET, .l_len = 0}; \
-  fcntl(fd, F_SETLK, lock); \
-  }
+#define FILE_LOCK_WRITE F_WRLCK
+#define FILE_LOCK_READ F_RDLCK
+
+static int
+file_lock_set(int fd, short cmd)
+{
+  struct flock lock = { .l_type = cmd, .l_start = 0, .l_whence = SEEK_SET, .l_len = 0};
+  return fcntl(fd, F_SETLKW, &lock);
+}
 #endif
 
 static void zathura_database_interface_init(ZathuraDatabaseInterface* iface);
 static void io_interface_init(GiraraInputHistoryIOInterface* iface);
 
+typedef struct zathura_plaindatabase_private_s {
+  char* bookmark_path;
+  GKeyFile* bookmarks;
+  GFileMonitor* bookmark_monitor;
+
+  char* history_path;
+  GKeyFile* history;
+  GFileMonitor* history_monitor;
+
+  char* input_history_path;
+} ZathuraPlainDatabasePrivate;
+
 G_DEFINE_TYPE_WITH_CODE(ZathuraPlainDatabase, zathura_plaindatabase, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE(ZATHURA_TYPE_DATABASE, zathura_database_interface_init)
-                        G_IMPLEMENT_INTERFACE(GIRARA_TYPE_INPUT_HISTORY_IO, io_interface_init))
+                        G_IMPLEMENT_INTERFACE(GIRARA_TYPE_INPUT_HISTORY_IO, io_interface_init)
+                        G_ADD_PRIVATE(ZathuraPlainDatabase))
 
 static void           plain_dispose(GObject* object);
 static void           plain_finalize(GObject* object);
@@ -71,21 +97,6 @@ static GKeyFile*      zathura_db_read_key_file_from_file(const char* path);
 static void           zathura_db_write_key_file_to_file(const char* file, GKeyFile* key_file);
 static void           cb_zathura_db_watch_file(GFileMonitor* monitor, GFile* file, GFile* other_file, GFileMonitorEvent event, zathura_database_t* database);
 
-typedef struct zathura_plaindatabase_private_s {
-  char* bookmark_path;
-  GKeyFile* bookmarks;
-  GFileMonitor* bookmark_monitor;
-
-  char* history_path;
-  GKeyFile* history;
-  GFileMonitor* history_monitor;
-
-  char* input_history_path;
-} zathura_plaindatabase_private_t;
-
-#define ZATHURA_PLAINDATABASE_GET_PRIVATE(obj) \
-  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), ZATHURA_TYPE_PLAINDATABASE, zathura_plaindatabase_private_t))
-
 enum {
   PROP_0,
   PROP_PATH
@@ -130,9 +141,6 @@ io_interface_init(GiraraInputHistoryIOInterface* iface)
 static void
 zathura_plaindatabase_class_init(ZathuraPlainDatabaseClass* class)
 {
-  /* add private members */
-  g_type_class_add_private(class, sizeof(zathura_plaindatabase_private_t));
-
   /* override methods */
   GObjectClass* object_class = G_OBJECT_CLASS(class);
   object_class->dispose      = plain_dispose;
@@ -147,7 +155,7 @@ zathura_plaindatabase_class_init(ZathuraPlainDatabaseClass* class)
 static void
 zathura_plaindatabase_init(ZathuraPlainDatabase* db)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(db);
 
   priv->bookmark_path         = NULL;
   priv->bookmark_monitor      = NULL;
@@ -164,7 +172,7 @@ zathura_plaindatabase_new(const char* path)
   g_return_val_if_fail(path != NULL && strlen(path) != 0, NULL);
 
   zathura_database_t* db = g_object_new(ZATHURA_TYPE_PLAINDATABASE, "path", path, NULL);
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(ZATHURA_PLAINDATABASE(db));
   if (priv->bookmark_path == NULL) {
     g_object_unref(db);
     return NULL;
@@ -176,7 +184,7 @@ zathura_plaindatabase_new(const char* path)
 static void
 plain_db_init(ZathuraPlainDatabase* db, const char* dir)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(db);
 
   /* bookmarks */
   priv->bookmark_path = g_build_filename(dir, BOOKMARKS, NULL);
@@ -287,7 +295,7 @@ static void
 plain_dispose(GObject* object)
 {
   ZathuraPlainDatabase* db = ZATHURA_PLAINDATABASE(object);
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(db);
 
   g_clear_object(&priv->bookmark_monitor);
   g_clear_object(&priv->history_monitor);
@@ -299,7 +307,7 @@ static void
 plain_finalize(GObject* object)
 {
   ZathuraPlainDatabase* db = ZATHURA_PLAINDATABASE(object);
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(db);
 
   /* bookmarks */
   g_free(priv->bookmark_path);
@@ -325,7 +333,8 @@ static bool
 plain_add_bookmark(zathura_database_t* db, const char* file,
                    zathura_bookmark_t* bookmark)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
   if (priv->bookmarks == NULL || priv->bookmark_path == NULL ||
       bookmark->id == NULL) {
     return false;
@@ -363,7 +372,8 @@ plain_add_bookmark(zathura_database_t* db, const char* file,
 static bool
 plain_remove_bookmark(zathura_database_t* db, const char* file, const char* id)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
   if (priv->bookmarks == NULL || priv->bookmark_path == NULL) {
     return false;
   }
@@ -386,7 +396,8 @@ plain_remove_bookmark(zathura_database_t* db, const char* file, const char* id)
 static girara_list_t*
 plain_load_bookmarks(zathura_database_t* db, const char* file)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
   if (priv->bookmarks == NULL) {
     return NULL;
   }
@@ -397,9 +408,7 @@ plain_load_bookmarks(zathura_database_t* db, const char* file)
     return NULL;
   }
 
-  girara_list_t* result = girara_sorted_list_new2((girara_compare_function_t)
-                          zathura_bookmarks_compare, (girara_free_function_t)
-                          zathura_bookmark_free);
+  girara_list_t* result = bookmarks_list_new();
 
   gsize num_keys;
   char** keys = g_key_file_get_keys(priv->bookmarks, name, &num_keys, NULL);
@@ -499,9 +508,10 @@ plain_load_jumplist(zathura_database_t* db, const char* file)
 {
   g_return_val_if_fail(db != NULL && file != NULL, NULL);
 
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
-  char* str_value = g_key_file_get_string(priv->history, file, KEY_JUMPLIST, NULL);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
 
+  char* str_value = g_key_file_get_string(priv->history, file, KEY_JUMPLIST, NULL);
   if (str_value == NULL) {
     return girara_list_new2(g_free);
   }
@@ -511,24 +521,31 @@ plain_load_jumplist(zathura_database_t* db, const char* file)
   return list;
 }
 
+static void
+jump_to_str(void* data, void* userdata)
+{
+  const zathura_jump_t* jump = data;
+  GString* str_val           = userdata;
+
+  char buffer[G_ASCII_DTOSTR_BUF_SIZE] = { '\0' };
+
+  g_string_append_printf(str_val, "%d ", jump->page);
+  g_string_append(str_val, g_ascii_dtostr(buffer, G_ASCII_DTOSTR_BUF_SIZE, jump->x));
+  g_string_append_c(str_val, ' ');
+  g_string_append(str_val, g_ascii_dtostr(buffer, G_ASCII_DTOSTR_BUF_SIZE, jump->y));
+  g_string_append_c(str_val, ' ');
+}
+
 static bool
 plain_save_jumplist(zathura_database_t* db, const char* file, girara_list_t* jumplist)
 {
   g_return_val_if_fail(db != NULL && file != NULL && jumplist != NULL, false);
 
   GString* str_val = g_string_new(NULL);
+  girara_list_foreach(jumplist, jump_to_str, str_val);
 
-  GIRARA_LIST_FOREACH(jumplist, zathura_jump_t*, iter, jump)
-    char buffer[G_ASCII_DTOSTR_BUF_SIZE] = { '\0' };
-
-    g_string_append_printf(str_val, "%d ", jump->page);
-    g_string_append(str_val, g_ascii_dtostr(buffer, G_ASCII_DTOSTR_BUF_SIZE, jump->x));
-    g_string_append_c(str_val, ' ');
-    g_string_append(str_val, g_ascii_dtostr(buffer, G_ASCII_DTOSTR_BUF_SIZE, jump->y));
-    g_string_append_c(str_val, ' ');
-  GIRARA_LIST_FOREACH_END(jumplist, zathura_jump_t*, iter, jump);
-
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
 
   g_key_file_set_string(priv->history, file, KEY_JUMPLIST, str_val->str);
   zathura_db_write_key_file_to_file(priv->history_path, priv->history);
@@ -541,7 +558,8 @@ static bool
 plain_set_fileinfo(zathura_database_t* db, const char* file, zathura_fileinfo_t*
                    file_info)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
   if (priv->history == NULL || file_info == NULL || file == NULL) {
     return false;
   }
@@ -550,10 +568,11 @@ plain_set_fileinfo(zathura_database_t* db, const char* file, zathura_fileinfo_t*
 
   g_key_file_set_integer(priv->history, name, KEY_PAGE,              file_info->current_page);
   g_key_file_set_integer(priv->history, name, KEY_OFFSET,            file_info->page_offset);
-  g_key_file_set_double (priv->history, name, KEY_SCALE,             file_info->scale);
+  g_key_file_set_double (priv->history, name, KEY_ZOOM,              file_info->zoom);
   g_key_file_set_integer(priv->history, name, KEY_ROTATE,            file_info->rotation);
   g_key_file_set_integer(priv->history, name, KEY_PAGES_PER_ROW,     file_info->pages_per_row);
   g_key_file_set_string(priv->history, name, KEY_FIRST_PAGE_COLUMN,  file_info->first_page_column_list);
+  g_key_file_set_boolean(priv->history, name, KEY_PAGE_RIGHT_TO_LEFT,file_info->page_right_to_left);
   g_key_file_set_double (priv->history, name, KEY_POSITION_X,        file_info->position_x);
   g_key_file_set_double (priv->history, name, KEY_POSITION_Y,        file_info->position_y);
   g_key_file_set_integer(priv->history, name, KEY_TIME,              time(NULL));
@@ -573,7 +592,8 @@ plain_get_fileinfo(zathura_database_t* db, const char* file, zathura_fileinfo_t*
     return false;
   }
 
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
   if (priv->history == NULL) {
     return false;
   }
@@ -586,7 +606,7 @@ plain_get_fileinfo(zathura_database_t* db, const char* file, zathura_fileinfo_t*
 
   file_info->current_page      = g_key_file_get_integer(priv->history, name, KEY_PAGE, NULL);
   file_info->page_offset       = g_key_file_get_integer(priv->history, name, KEY_OFFSET, NULL);
-  file_info->scale             = g_key_file_get_double (priv->history, name, KEY_SCALE, NULL);
+  file_info->zoom              = g_key_file_get_double (priv->history, name, KEY_ZOOM, NULL);
   file_info->rotation          = g_key_file_get_integer(priv->history, name, KEY_ROTATE, NULL);
 
   /* the following flags got introduced at a later point */
@@ -596,6 +616,9 @@ plain_get_fileinfo(zathura_database_t* db, const char* file, zathura_fileinfo_t*
   if (g_key_file_has_key(priv->history, name, KEY_FIRST_PAGE_COLUMN, NULL) == TRUE) {
     file_info->first_page_column_list = g_key_file_get_string(priv->history, name, KEY_FIRST_PAGE_COLUMN, NULL);
   }
+  if (g_key_file_has_key(priv->history, name, KEY_PAGE_RIGHT_TO_LEFT, NULL) == TRUE) {
+    file_info->page_right_to_left = g_key_file_get_boolean(priv->history, name, KEY_PAGE_RIGHT_TO_LEFT, NULL);
+  }
   if (g_key_file_has_key(priv->history, name, KEY_POSITION_X, NULL) == TRUE) {
     file_info->position_x        = g_key_file_get_double(priv->history, name, KEY_POSITION_X, NULL);
   }
@@ -637,10 +660,15 @@ zathura_db_read_key_file_from_file(const char* path)
   }
 
   /* open file */
-  FILE* file = fopen(path, "rw");
+  FILE* file = fopen(path, "r+");
   if (file == NULL) {
     return NULL;
   }
+  /* and lock it */
+  if (file_lock_set(fileno(file), FILE_LOCK_WRITE) != 0) {
+    fclose(file);
+    return NULL;
+  }
 
   GKeyFile* key_file = g_key_file_new();
   if (key_file == NULL) {
@@ -649,9 +677,7 @@ zathura_db_read_key_file_from_file(const char* path)
   }
 
   /* read config file */
-  file_lock_set(fileno(file), F_WRLCK);
   char* content = girara_file_read2(file);
-  file_lock_set(fileno(file), F_UNLCK);
   fclose(file);
   if (content == NULL) {
     g_key_file_free(key_file);
@@ -713,12 +739,9 @@ zathura_db_write_key_file_to_file(const char* file, GKeyFile* key_file)
     return;
   }
 
-  file_lock_set(fd, F_WRLCK);
-  if (write(fd, content, strlen(content)) == 0) {
+  if (file_lock_set(fd, FILE_LOCK_READ) != 0 || write(fd, content, strlen(content)) == 0) {
     girara_error("Failed to write to %s", file);
   }
-  file_lock_set(fd, F_UNLCK);
-
   close(fd);
 
   g_free(content);
@@ -737,7 +760,8 @@ cb_zathura_db_watch_file(GFileMonitor* UNUSED(monitor), GFile* file, GFile* UNUS
     return;
   }
 
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(database);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(database);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
   if (priv->bookmark_path && strcmp(priv->bookmark_path, path) == 0) {
     if (priv->bookmarks != NULL) {
       g_key_file_free(priv->bookmarks);
@@ -758,7 +782,8 @@ cb_zathura_db_watch_file(GFileMonitor* UNUSED(monitor), GFile* file, GFile* UNUS
 static girara_list_t*
 plain_io_read(GiraraInputHistoryIO* db)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
 
   /* open file */
   FILE* file = fopen(priv->input_history_path, "r");
@@ -767,9 +792,11 @@ plain_io_read(GiraraInputHistoryIO* db)
   }
 
   /* read input history file */
-  file_lock_set(fileno(file), F_RDLCK);
+  if (file_lock_set(fileno(file), FILE_LOCK_READ) != 0) {
+    fclose(file);
+    return NULL;
+  }
   char* content = girara_file_read2(file);
-  file_lock_set(fileno(file), F_UNLCK);
   fclose(file);
 
   girara_list_t* res = girara_list_new2(g_free);
@@ -789,7 +816,8 @@ plain_io_read(GiraraInputHistoryIO* db)
 static void
 plain_io_append(GiraraInputHistoryIO* db, const char* input)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
 
   /* open file */
   FILE* file = fopen(priv->input_history_path, "r+");
@@ -798,13 +826,15 @@ plain_io_append(GiraraInputHistoryIO* db, const char* input)
   }
 
   /* read input history file */
-  file_lock_set(fileno(file), F_WRLCK);
+  if (file_lock_set(fileno(file), FILE_LOCK_WRITE) != 0) {
+    fclose(file);
+    return;
+  }
   char* content = girara_file_read2(file);
 
   rewind(file);
   if (ftruncate(fileno(file), 0) != 0) {
     free(content);
-    file_lock_set(fileno(file), F_UNLCK);
     fclose(file);
     return;
   }
@@ -821,8 +851,6 @@ plain_io_append(GiraraInputHistoryIO* db, const char* input)
   }
   g_strfreev(tmp);
   fprintf(file, "%s\n", input);
-
-  file_lock_set(fileno(file), F_UNLCK);
   fclose(file);
 }
 
@@ -854,7 +882,8 @@ compare_time(const void* l, const void* r, void* data)
 static girara_list_t*
 plain_get_recent_files(zathura_database_t* db, int max, const char* basepath)
 {
-  zathura_plaindatabase_private_t* priv = ZATHURA_PLAINDATABASE_GET_PRIVATE(db);
+  ZathuraPlainDatabase* plaindb     = ZATHURA_PLAINDATABASE(db);
+  ZathuraPlainDatabasePrivate* priv = zathura_plaindatabase_get_instance_private(plaindb);
 
   girara_list_t* result = girara_list_new2(g_free);
   if (result == NULL) {
