@@ -1,4 +1,4 @@
-/* See LICENSE file for license and copyright information */
+/* SPDX-License-Identifier: Zlib */
 
 #include <sqlite3.h>
 #include <girara/utils.h>
@@ -10,12 +10,26 @@
 #include "database-sqlite.h"
 #include "utils.h"
 
+/* version of the database layout */
+#define DATABASE_VERSION 1
+
+static char*
+sqlite3_column_text_dup(sqlite3_stmt* stmt, int col)
+{
+  return g_strdup((const char*) sqlite3_column_text(stmt, col));
+}
+
 static void zathura_database_interface_init(ZathuraDatabaseInterface* iface);
 static void io_interface_init(GiraraInputHistoryIOInterface* iface);
 
+typedef struct zathura_sqldatabase_private_s {
+  sqlite3* session;
+} ZathuraSQLDatabasePrivate;
+
 G_DEFINE_TYPE_WITH_CODE(ZathuraSQLDatabase, zathura_sqldatabase, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE(ZATHURA_TYPE_DATABASE, zathura_database_interface_init)
-                        G_IMPLEMENT_INTERFACE(GIRARA_TYPE_INPUT_HISTORY_IO, io_interface_init))
+                        G_IMPLEMENT_INTERFACE(GIRARA_TYPE_INPUT_HISTORY_IO, io_interface_init)
+                        G_ADD_PRIVATE(ZathuraSQLDatabase))
 
 static bool           check_column(sqlite3* session, const char* table, const char* col, bool* result);
 static bool           check_column_type(sqlite3* session, const char* table, const char* col, const char* type, bool* result);
@@ -32,13 +46,6 @@ static void           sqlite_io_append(GiraraInputHistoryIO* db, const char*);
 static girara_list_t* sqlite_io_read(GiraraInputHistoryIO* db);
 static girara_list_t* sqlite_get_recent_files(zathura_database_t* db, int max, const char* basepath);
 
-typedef struct zathura_sqldatabase_private_s {
-  sqlite3* session;
-} zathura_sqldatabase_private_t;
-
-#define ZATHURA_SQLDATABASE_GET_PRIVATE(obj) \
-  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), ZATHURA_TYPE_SQLDATABASE, zathura_sqldatabase_private_t))
-
 enum {
   PROP_0,
   PROP_PATH
@@ -69,9 +76,6 @@ io_interface_init(GiraraInputHistoryIOInterface* iface)
 static void
 zathura_sqldatabase_class_init(ZathuraSQLDatabaseClass* class)
 {
-  /* add private members */
-  g_type_class_add_private(class, sizeof(zathura_sqldatabase_private_t));
-
   /* override methods */
   GObjectClass* object_class = G_OBJECT_CLASS(class);
   object_class->finalize     = sqlite_finalize;
@@ -85,7 +89,7 @@ zathura_sqldatabase_class_init(ZathuraSQLDatabaseClass* class)
 static void
 zathura_sqldatabase_init(ZathuraSQLDatabase* db)
 {
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(db);
   priv->session = NULL;
 }
 
@@ -95,7 +99,7 @@ zathura_sqldatabase_new(const char* path)
   g_return_val_if_fail(path != NULL && strlen(path) != 0, NULL);
 
   zathura_database_t* db = g_object_new(ZATHURA_TYPE_SQLDATABASE, "path", path, NULL);
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(ZATHURA_SQLDATABASE(db));
   if (priv->session == NULL) {
     g_object_unref(G_OBJECT(db));
     return NULL;
@@ -108,19 +112,57 @@ static void
 sqlite_finalize(GObject* object)
 {
   ZathuraSQLDatabase* db = ZATHURA_SQLDATABASE(object);
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(db);
   if (priv->session) {
+    sqlite3_exec(priv->session, "VACUUM;", NULL, 0, NULL);
     sqlite3_close(priv->session);
   }
 
   G_OBJECT_CLASS(zathura_sqldatabase_parent_class)->finalize(object);
 }
 
-static void
-sqlite_db_init(ZathuraSQLDatabase* db, const char* path)
+static sqlite3_stmt*
+prepare_statement(sqlite3* session, const char* statement)
 {
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  if (session == NULL || statement == NULL) {
+    return NULL;
+  }
 
+  const char* pz_tail   = NULL;
+  sqlite3_stmt* pp_stmt = NULL;
+
+  if (sqlite3_prepare_v2(session, statement, -1, &pp_stmt, &pz_tail) != SQLITE_OK) {
+    girara_error("Failed to prepare query: %s", statement);
+    sqlite3_finalize(pp_stmt);
+    return NULL;
+  } else if (pz_tail && *pz_tail != '\0') {
+    girara_error("Unused portion of statement: %s", pz_tail);
+    sqlite3_finalize(pp_stmt);
+    return NULL;
+  }
+
+  return pp_stmt;
+}
+
+static int
+sqlite_get_user_version(sqlite3* session)
+{
+  sqlite3_stmt* stmt = prepare_statement(session, "PRAGMA user_version;");
+  if (stmt == NULL) {
+    return -1;
+  }
+
+  int version = -1;
+  if (sqlite3_step(stmt) == SQLITE_ROW) {
+    version = sqlite3_column_int(stmt, 0);
+  }
+  sqlite3_finalize(stmt);
+  return version;
+}
+
+static void
+sqlite_db_check_layout(sqlite3* session, const int database_version, const bool new_db)
+{
   /* create bookmarks table */
   static const char SQL_BOOKMARK_INIT[] =
     "CREATE TABLE IF NOT EXISTS bookmarks ("
@@ -131,7 +173,7 @@ sqlite_db_init(ZathuraSQLDatabase* db, const char* path)
     "vadj_ratio FLOAT,"
     "PRIMARY KEY(file, id));";
 
-  /* ceate jumplist table */
+  /* create jumplist table */
   static const char SQL_JUMPLIST_INIT[] =
     "CREATE TABLE IF NOT EXISTS jumplist ("
     "id INTEGER PRIMARY KEY AUTOINCREMENT,"
@@ -147,13 +189,14 @@ sqlite_db_init(ZathuraSQLDatabase* db, const char* path)
     "file TEXT PRIMARY KEY,"
     "page INTEGER,"
     "offset INTEGER,"
-    "scale FLOAT,"
+    "zoom FLOAT,"
     "rotation INTEGER,"
     "pages_per_row INTEGER,"
     "first_page_column TEXT,"
     "position_x FLOAT,"
     "position_y FLOAT,"
-    "time TIMESTAMP"
+    "time TIMESTAMP,"
+    "page_right_to_left INTEGER"
     ");";
 
   /* create history table */
@@ -184,90 +227,154 @@ sqlite_db_init(ZathuraSQLDatabase* db, const char* path)
   static const char SQL_FILEINFO_ALTER3[] =
     "ALTER TABLE fileinfo ADD COLUMN time TIMESTAMP;";
 
+  /* update fileinfo table (part 4) */
+  static const char SQL_FILEINFO_ALTER4[] =
+    "ALTER TABLE fileinfo ADD COLUMN zoom FLOAT;";
+
+  /* update fileinfo table (part 5) */
+  static const char SQL_FILEINFO_ALTER5[] =
+    "ALTER TABLE fileinfo ADD COLUMN page_right_to_left INTEGER;";
+
   /* update bookmark table */
   static const char SQL_BOOKMARK_ALTER[] =
     "ALTER TABLE bookmarks ADD COLUMN hadj_ratio FLOAT;"
     "ALTER TABLE bookmarks ADD COLUMN vadj_ratio FLOAT;";
 
+  /* create tables if they don't exist */
+  for (size_t s = 0; s < LENGTH(ALL_INIT); ++s) {
+    if (sqlite3_exec(session, ALL_INIT[s], NULL, 0, NULL) != SQLITE_OK) {
+      girara_error("Failed to initialize database");
+      sqlite3_close(session);
+      return;
+    }
+  }
+  if (new_db == true)
+    return;
+
+#ifndef SQLITE_OMIT_COMPILEOPTION_DIAGS
+  if (sqlite3_compileoption_used("SQLITE_OMIT_ALTERTABLE") == 1) {
+    girara_error("sqlite3 built without support for ALTER, cannot update database");
+    return;
+  }
+#endif
+
+  bool all_updates_ok = true;
+  if (database_version < 1)
+  {
+    /* check existing tables for missing columns */
+    bool res1, ret1;
+    ret1 = check_column(session, "fileinfo", "pages_per_row", &res1);
+
+    if (ret1 == true && res1 == false) {
+      if (sqlite3_exec(session, SQL_FILEINFO_ALTER, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: pages_per_row, position_x, position_y");
+        all_updates_ok = false;
+      }
+    }
+
+    ret1 = check_column(session, "fileinfo", "first_page_column", &res1);
+
+    if (ret1 == true && res1 == false) {
+      if (sqlite3_exec(session, SQL_FILEINFO_ALTER2, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: first_page_column");
+        all_updates_ok = false;
+      }
+    }
+
+    ret1 = check_column(session, "fileinfo", "time", &res1);
+
+    if (ret1 == true && res1 == false) {
+      if (sqlite3_exec(session, SQL_FILEINFO_ALTER3, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: time");
+        all_updates_ok = false;
+      }
+    }
+
+    ret1 = check_column(session, "fileinfo", "zoom", &res1);
+
+    if (ret1 == true && res1 == false) {
+      if (sqlite3_exec(session, SQL_FILEINFO_ALTER4, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: zoom");
+        all_updates_ok = false;
+      }
+    }
+
+    ret1 = check_column(session, "fileinfo", "page_right_to_left", &res1);
+
+    if (ret1 == true && res1 == false) {
+      if (sqlite3_exec(session, SQL_FILEINFO_ALTER5, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: pages_right_to_left");
+        all_updates_ok = false;
+      }
+    }
+
+    ret1 = check_column(session, "bookmarks", "hadj_ratio", &res1);
+
+    if (ret1 == true && res1 == false) {
+      if (sqlite3_exec(session, SQL_BOOKMARK_ALTER, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: hadj_ration, vadj_ratio");
+        all_updates_ok = false;
+      }
+    }
+
+    /* check existing tables for correct column types */
+    ret1 = check_column_type(session, "fileinfo", "first_page_column", "TEXT", &res1);
+
+    if (ret1 == true && res1 == false) {
+      /* prepare transaction */
+      static const char tx_begin[] =
+        "BEGIN TRANSACTION;"
+        "ALTER TABLE fileinfo RENAME TO tmp;";
+      static const char tx_end[] =
+        "INSERT INTO fileinfo SELECT * FROM tmp;"
+        "DROP TABLE tmp;"
+        "COMMIT;";
+
+      /* assemble transaction */
+      char transaction[sizeof(tx_begin) + sizeof(SQL_FILEINFO_INIT) + sizeof(tx_end) - 2] = { '\0' };
+      g_strlcat(transaction, tx_begin, sizeof(transaction));
+      g_strlcat(transaction, SQL_FILEINFO_INIT, sizeof(transaction));
+      g_strlcat(transaction, tx_end, sizeof(transaction));
+
+      if (sqlite3_exec(session, transaction, NULL, 0, NULL) != SQLITE_OK) {
+        girara_warning("failed to update database table layout: first_page_column");
+        all_updates_ok = false;
+      }
+    }
+  }
+
+  /* update database version if all updates were successful */
+  if (all_updates_ok == true) {
+    sqlite3_exec(session, "PRAGMA user_version = " G_STRINGIFY(DATABASE_VERSION) ";", NULL, 0, NULL);
+  }
+}
+
+static void
+sqlite_db_init(ZathuraSQLDatabase* db, const char* path)
+{
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(db);
+
+  const bool db_exists = g_file_test(path, G_FILE_TEST_EXISTS);
   sqlite3* session = NULL;
   if (sqlite3_open(path, &session) != SQLITE_OK) {
     girara_error("Could not open database: %s\n", path);
     return;
   }
 
-  /* create tables if they don't exist */
-  for (size_t s = 0; s < LENGTH(ALL_INIT); ++s) {
-    if (sqlite3_exec(session, ALL_INIT[s], NULL, 0, NULL) != SQLITE_OK) {
-      girara_error("Failed to initialize database: %s\n", path);
-      sqlite3_close(session);
-      return;
-    }
+  /* Set busy timeout to 1s. */
+  sqlite3_busy_timeout(session, 1000);
+
+  const int database_version = sqlite_get_user_version(session);
+  if (database_version == -1) {
+    girara_error("Failed to query database version.");
+    sqlite3_close(session);
+    return;
   }
 
-  /* check existing tables for missing columns */
-  bool res1, res2, ret1, ret2;
-
-  ret1 = check_column(session, "fileinfo", "pages_per_row", &res1);
-
-  if (ret1 == true && res1 == false) {
+  girara_debug("database version: %d (current: %d)", database_version, DATABASE_VERSION);
+  if (database_version < DATABASE_VERSION) {
     girara_debug("old database table layout detected; updating ...");
-    if (sqlite3_exec(session, SQL_FILEINFO_ALTER, NULL, 0, NULL) != SQLITE_OK) {
-      girara_warning("failed to update database table layout");
-    }
-  }
-
-  ret1 = check_column(session, "fileinfo", "first_page_column", &res1);
-
-  if (ret1 == true && res1 == false) {
-    girara_debug("old database table layout detected; updating ...");
-    if (sqlite3_exec(session, SQL_FILEINFO_ALTER2, NULL, 0, NULL) != SQLITE_OK) {
-      girara_warning("failed to update database table layout");
-    }
-  }
-
-  ret1 = check_column(session, "fileinfo", "time", &res1);
-
-  if (ret1 == true && res1 == false) {
-    girara_debug("old database table layout detected; updating ...");
-    if (sqlite3_exec(session, SQL_FILEINFO_ALTER3, NULL, 0, NULL) != SQLITE_OK) {
-      girara_warning("failed to update database table layout");
-    }
-  }
-
-  ret1 = check_column(session, "bookmarks", "hadj_ratio", &res1);
-  ret2 = check_column(session, "bookmarks", "vadj_ratio", &res2);
-
-  if (ret1 == true && ret2 == true && res1 == false && res2 == false) {
-    girara_debug("old database table layout detected; updating ...");
-    if (sqlite3_exec(session, SQL_BOOKMARK_ALTER, NULL, 0, NULL) != SQLITE_OK) {
-      girara_warning("failed to update database table layout");
-    }
-  }
-
-  /* check existing tables for correct column types */
-  ret1 = check_column_type(session, "fileinfo", "first_page_column", "TEXT", &res1);
-
-  if (ret1 == true && res1 == false) {
-    girara_debug("old database table layout detected; updating ...");
-
-    /* prepare transaction */
-    static const char tx_begin[] =
-      "BEGIN TRANSACTION;"
-      "ALTER TABLE fileinfo RENAME TO tmp;";
-    static const char tx_end[] =
-      "INSERT INTO fileinfo SELECT * FROM tmp;"
-      "DROP TABLE tmp;"
-      "COMMIT;";
-
-    /* assemble transaction */
-    char transaction[sizeof(tx_begin) + sizeof(SQL_FILEINFO_INIT) + sizeof(tx_end) - 2] = { '\0' };
-    g_strlcat(transaction, tx_begin, sizeof(transaction));
-    g_strlcat(transaction, SQL_FILEINFO_INIT, sizeof(transaction));
-    g_strlcat(transaction, tx_end, sizeof(transaction));
-
-    if (sqlite3_exec(session, transaction, NULL, 0, NULL) != SQLITE_OK) {
-      girara_warning("failed to update database table layout");
-    }
+    sqlite_db_check_layout(session, database_version, !db_exists);
   }
 
   priv->session = session;
@@ -276,8 +383,8 @@ sqlite_db_init(ZathuraSQLDatabase* db, const char* path)
 static void
 sqlite_set_property(GObject* object, guint prop_id, const GValue* value, GParamSpec* pspec)
 {
-  ZathuraSQLDatabase* db = ZATHURA_SQLDATABASE(object);
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* db          = ZATHURA_SQLDATABASE(object);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(db);
 
   switch (prop_id) {
     case PROP_PATH:
@@ -289,29 +396,6 @@ sqlite_set_property(GObject* object, guint prop_id, const GValue* value, GParamS
   }
 }
 
-static sqlite3_stmt*
-prepare_statement(sqlite3* session, const char* statement)
-{
-  if (session == NULL || statement == NULL) {
-    return NULL;
-  }
-
-  const char* pz_tail   = NULL;
-  sqlite3_stmt* pp_stmt = NULL;
-
-  if (sqlite3_prepare_v2(session, statement, -1, &pp_stmt, &pz_tail) != SQLITE_OK) {
-    girara_error("Failed to prepare query: %s", statement);
-    sqlite3_finalize(pp_stmt);
-    return NULL;
-  } else if (pz_tail && *pz_tail != '\0') {
-    girara_error("Unused portion of statement: %s", pz_tail);
-    sqlite3_finalize(pp_stmt);
-    return NULL;
-  }
-
-  return pp_stmt;
-}
-
 static bool
 check_column(sqlite3* session, const char* table, const char* col, bool* res)
 {
@@ -337,7 +421,7 @@ check_column(sqlite3* session, const char* table, const char* col, bool* res)
   }
 
   if (*res == false) {
-    girara_debug("column %s in table %s is NOT found", col, table);
+    girara_debug("Column '%s' in table '%s' NOT found.", col, table);
   }
 
   sqlite3_finalize(stmt);
@@ -373,7 +457,7 @@ check_column_type(sqlite3* session, const char* table, const char* col, const ch
   }
 
   if (*res == false) {
-    girara_debug("column %s in table %s has wrong type", col, table);
+    girara_debug("Column '%s' in table '%s' has wrong type.", col, table);
   }
 
   sqlite3_finalize(stmt);
@@ -386,7 +470,8 @@ static bool
 sqlite_add_bookmark(zathura_database_t* db, const char* file,
                     zathura_bookmark_t* bookmark)
 {
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
 
   static const char SQL_BOOKMARK_ADD[] =
     "REPLACE INTO bookmarks (file, id, page, hadj_ratio, vadj_ratio) VALUES (?, ?, ?, ?, ?);";
@@ -401,7 +486,6 @@ sqlite_add_bookmark(zathura_database_t* db, const char* file,
       sqlite3_bind_int(stmt, 3, bookmark->page) != SQLITE_OK ||
       sqlite3_bind_double(stmt, 4, bookmark->x) != SQLITE_OK ||
       sqlite3_bind_double(stmt, 5, bookmark->y) != SQLITE_OK) {
-
     sqlite3_finalize(stmt);
     girara_error("Failed to bind arguments.");
     return false;
@@ -417,7 +501,8 @@ static bool
 sqlite_remove_bookmark(zathura_database_t* db, const char* file, const char*
                        id)
 {
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb      = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
 
   static const char SQL_BOOKMARK_ADD[] =
     "DELETE FROM bookmarks WHERE file = ? AND id = ?;";
@@ -443,7 +528,8 @@ sqlite_remove_bookmark(zathura_database_t* db, const char* file, const char*
 static girara_list_t*
 sqlite_load_bookmarks(zathura_database_t* db, const char* file)
 {
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
 
   static const char SQL_BOOKMARK_SELECT[] =
     "SELECT id, page, hadj_ratio, vadj_ratio FROM bookmarks WHERE file = ?;";
@@ -459,8 +545,7 @@ sqlite_load_bookmarks(zathura_database_t* db, const char* file)
     return NULL;
   }
 
-  girara_list_t* result = girara_sorted_list_new2((girara_compare_function_t) zathura_bookmarks_compare,
-                          (girara_free_function_t) zathura_bookmark_free);
+  girara_list_t* result = bookmarks_list_new();
   if (result == NULL) {
     sqlite3_finalize(stmt);
     return NULL;
@@ -472,12 +557,12 @@ sqlite_load_bookmarks(zathura_database_t* db, const char* file)
       continue;
     }
 
-    bookmark->id   = g_strdup((const char*) sqlite3_column_text(stmt, 0));
+    bookmark->id   = sqlite3_column_text_dup(stmt, 0);
     bookmark->page = sqlite3_column_int(stmt, 1);
     bookmark->x    = sqlite3_column_double(stmt, 2);
     bookmark->y    = sqlite3_column_double(stmt, 3);
-    bookmark->x    = bookmark->x <= 0.0 ? DBL_MIN : bookmark->x;
-    bookmark->y    = bookmark->y <= 0.0 ? DBL_MIN : bookmark->y;
+    bookmark->x    = MAX(DBL_MIN, bookmark->x);
+    bookmark->y    = MAX(DBL_MIN, bookmark->y);
 
     girara_list_append(result, bookmark);
   }
@@ -492,11 +577,14 @@ sqlite_save_jumplist(zathura_database_t* db, const char* file, girara_list_t* ju
 {
   g_return_val_if_fail(db != NULL && file != NULL && jumplist != NULL, false);
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
-  static const char SQL_INSERT_JUMP[] = "INSERT INTO jumplist (file, page, hadj_ratio, vadj_ratio) VALUES (?, ?, ?, ?);";
+  static const char SQL_INSERT_JUMP[]     = "INSERT INTO jumplist (file, page, hadj_ratio, vadj_ratio) VALUES (?, ?, ?, ?);";
   static const char SQL_REMOVE_JUMPLIST[] = "DELETE FROM jumplist WHERE file = ?;";
+
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
+
   sqlite3_stmt* stmt = NULL;
-  int res = 0;
+  int res            = 0;
 
   if (sqlite3_exec(priv->session, "BEGIN;", NULL, 0, NULL) != SQLITE_OK) {
     return false;
@@ -529,14 +617,9 @@ sqlite_save_jumplist(zathura_database_t* db, const char* file, girara_list_t* ju
     return true;
   }
 
-  girara_list_iterator_t* cur = girara_list_iterator(jumplist);
   bool status = true;
-
-  while (true) {
-    zathura_jump_t* jump = girara_list_iterator_data(cur);
-
+  GIRARA_LIST_FOREACH_BODY(jumplist, zathura_jump_t*, jump,
     stmt = prepare_statement(priv->session, SQL_INSERT_JUMP);
-
     if (stmt == NULL) {
       status = false;
       break;
@@ -546,7 +629,6 @@ sqlite_save_jumplist(zathura_database_t* db, const char* file, girara_list_t* ju
         sqlite3_bind_int(stmt, 2, jump->page)      != SQLITE_OK ||
         sqlite3_bind_double(stmt, 3, jump->x)      != SQLITE_OK ||
         sqlite3_bind_double(stmt, 4, jump->y)      != SQLITE_OK) {
-
       sqlite3_finalize(stmt);
       girara_error("Failed to bind arguments.");
       status = false;
@@ -560,15 +642,7 @@ sqlite_save_jumplist(zathura_database_t* db, const char* file, girara_list_t* ju
       status = false;
       break;
     }
-
-    if (girara_list_iterator_has_next(cur) == true) {
-      girara_list_iterator_next(cur);
-    } else {
-      break;
-    }
-  }
-
-  girara_list_iterator_free(cur);
+  );
 
   if (status == false) {
     sqlite3_exec(priv->session, "ROLLBACK;", NULL, 0, NULL);
@@ -584,10 +658,12 @@ sqlite_load_jumplist(zathura_database_t* db, const char* file)
 {
   g_return_val_if_fail(db != NULL && file != NULL, NULL);
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
   static const char SQL_GET_JUMPLIST[] = "SELECT page, hadj_ratio, vadj_ratio FROM jumplist WHERE file = ? ORDER BY id ASC;";
-  sqlite3_stmt* stmt = prepare_statement(priv->session, SQL_GET_JUMPLIST);
 
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
+
+  sqlite3_stmt* stmt = prepare_statement(priv->session, SQL_GET_JUMPLIST);
   if (stmt == NULL) {
     return NULL;
   }
@@ -637,26 +713,28 @@ sqlite_set_fileinfo(zathura_database_t* db, const char* file,
     return false;
   }
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
 
   static const char SQL_FILEINFO_SET[] =
-    "REPLACE INTO fileinfo (file, page, offset, scale, rotation, pages_per_row, first_page_column, position_x, position_y, time) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, DATETIME('now'));";
+    "REPLACE INTO fileinfo (file, page, offset, zoom, rotation, pages_per_row, first_page_column, position_x, position_y, time, page_right_to_left) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, DATETIME('now'), ?);";
 
   sqlite3_stmt* stmt = prepare_statement(priv->session, SQL_FILEINFO_SET);
   if (stmt == NULL) {
     return false;
   }
 
-  if (sqlite3_bind_text(stmt,   1, file, -1, NULL)               != SQLITE_OK ||
-      sqlite3_bind_int(stmt,    2, file_info->current_page)      != SQLITE_OK ||
-      sqlite3_bind_int(stmt,    3, file_info->page_offset)       != SQLITE_OK ||
-      sqlite3_bind_double(stmt, 4, file_info->scale)             != SQLITE_OK ||
-      sqlite3_bind_int(stmt,    5, file_info->rotation)          != SQLITE_OK ||
-      sqlite3_bind_int(stmt,    6, file_info->pages_per_row)     != SQLITE_OK ||
-      sqlite3_bind_text(stmt,   7, file_info->first_page_column_list, -1, NULL)
-                                                                 != SQLITE_OK ||
-      sqlite3_bind_double(stmt, 8, file_info->position_x)        != SQLITE_OK ||
-      sqlite3_bind_double(stmt, 9, file_info->position_y)        != SQLITE_OK) {
+  if (sqlite3_bind_text(stmt,    1, file, -1, NULL)                != SQLITE_OK ||
+      sqlite3_bind_int(stmt,     2, file_info->current_page)       != SQLITE_OK ||
+      sqlite3_bind_int(stmt,     3, file_info->page_offset)        != SQLITE_OK ||
+      sqlite3_bind_double(stmt,  4, file_info->zoom)               != SQLITE_OK ||
+      sqlite3_bind_int(stmt,     5, file_info->rotation)           != SQLITE_OK ||
+      sqlite3_bind_int(stmt,     6, file_info->pages_per_row)      != SQLITE_OK ||
+      sqlite3_bind_text(stmt,    7, file_info->first_page_column_list, -1, NULL)
+                                                                   != SQLITE_OK ||
+      sqlite3_bind_double(stmt,  8, file_info->position_x)         != SQLITE_OK ||
+      sqlite3_bind_double(stmt,  9, file_info->position_y)         != SQLITE_OK ||
+      sqlite3_bind_int(stmt,    10, file_info->page_right_to_left) != SQLITE_OK) {
     sqlite3_finalize(stmt);
     girara_error("Failed to bind arguments.");
     return false;
@@ -676,10 +754,11 @@ sqlite_get_fileinfo(zathura_database_t* db, const char* file,
     return false;
   }
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
 
   static const char SQL_FILEINFO_GET[] =
-    "SELECT page, offset, scale, rotation, pages_per_row, first_page_column, position_x, position_y FROM fileinfo WHERE file = ?;";
+    "SELECT page, offset, zoom, rotation, pages_per_row, first_page_column, position_x, position_y, page_right_to_left FROM fileinfo WHERE file = ?;";
 
   sqlite3_stmt* stmt = prepare_statement(priv->session, SQL_FILEINFO_GET);
   if (stmt == NULL) {
@@ -700,12 +779,13 @@ sqlite_get_fileinfo(zathura_database_t* db, const char* file,
 
   file_info->current_page           = sqlite3_column_int(stmt, 0);
   file_info->page_offset            = sqlite3_column_int(stmt, 1);
-  file_info->scale                  = sqlite3_column_double(stmt, 2);
+  file_info->zoom                   = sqlite3_column_double(stmt, 2);
   file_info->rotation               = sqlite3_column_int(stmt, 3);
   file_info->pages_per_row          = sqlite3_column_int(stmt, 4);
-  file_info->first_page_column_list = g_strdup((const char*) sqlite3_column_text(stmt, 5));
+  file_info->first_page_column_list = sqlite3_column_text_dup(stmt, 5);
   file_info->position_x             = sqlite3_column_double(stmt, 6);
   file_info->position_y             = sqlite3_column_double(stmt, 7);
+  file_info->page_right_to_left     = sqlite3_column_int(stmt, 8) != 0;
 
   sqlite3_finalize(stmt);
 
@@ -718,7 +798,9 @@ sqlite_io_append(GiraraInputHistoryIO* db, const char* input)
   static const char SQL_HISTORY_SET[] =
     "REPLACE INTO history (line, time) VALUES (?, DATETIME('now'));";
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
+
   sqlite3_stmt* stmt = prepare_statement(priv->session, SQL_HISTORY_SET);
   if (stmt == NULL) {
     return;
@@ -740,20 +822,22 @@ sqlite_io_read(GiraraInputHistoryIO* db)
   static const char SQL_HISTORY_GET[] =
     "SELECT line FROM history ORDER BY time";
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
+
   sqlite3_stmt* stmt = prepare_statement(priv->session, SQL_HISTORY_GET);
   if (stmt == NULL) {
     return NULL;
   }
 
-  girara_list_t* list = girara_list_new2((girara_free_function_t) g_free);
+  girara_list_t* list = girara_list_new2(g_free);
   if (list == NULL) {
     sqlite3_finalize(stmt);
     return NULL;
   }
 
   while (sqlite3_step(stmt) == SQLITE_ROW) {
-    girara_list_append(list, g_strdup((const char*) sqlite3_column_text(stmt, 0)));
+    girara_list_append(list, sqlite3_column_text_dup(stmt, 0));
   }
 
   sqlite3_finalize(stmt);
@@ -768,7 +852,9 @@ sqlite_get_recent_files(zathura_database_t* db, int max, const char* basepath)
   static const char SQL_HISTORY_GET_WITH_BASEPATH[] =
     "SELECT file FROM fileinfo WHERE file LIKE ? || '%' ORDER BY time DESC LIMIT ?";
 
-  zathura_sqldatabase_private_t* priv = ZATHURA_SQLDATABASE_GET_PRIVATE(db);
+  ZathuraSQLDatabase* sqldb       = ZATHURA_SQLDATABASE(db);
+  ZathuraSQLDatabasePrivate* priv = zathura_sqldatabase_get_instance_private(sqldb);
+
   sqlite3_stmt* stmt = prepare_statement(priv->session, basepath == NULL ? SQL_HISTORY_GET : SQL_HISTORY_GET_WITH_BASEPATH);
   if (stmt == NULL) {
     return NULL;
@@ -791,14 +877,14 @@ sqlite_get_recent_files(zathura_database_t* db, int max, const char* basepath)
     return false;
   }
 
-  girara_list_t* list = girara_list_new2((girara_free_function_t) g_free);
+  girara_list_t* list = girara_list_new2(g_free);
   if (list == NULL) {
     sqlite3_finalize(stmt);
     return NULL;
   }
 
   while (sqlite3_step(stmt) == SQLITE_ROW) {
-    girara_list_append(list, g_strdup((const char*) sqlite3_column_text(stmt, 0)));
+    girara_list_append(list, sqlite3_column_text_dup(stmt, 0));
   }
 
   sqlite3_finalize(stmt);
