@@ -1,7 +1,4 @@
-/* See LICENSE file for license and copyright information */
-
-#define _DEFAULT_SOURCE
-#define _XOPEN_SOURCE 700
+/* SPDX-License-Identifier: Zlib */
 
 #include <errno.h>
 #include <stdlib.h>
@@ -18,6 +15,10 @@
 #include <glib/gstdio.h>
 #include <glib/gi18n.h>
 
+#ifdef GDK_WINDOWING_WAYLAND
+#include <gdk/gdkwayland.h>
+#endif
+
 #ifdef G_OS_UNIX
 #include <glib-unix.h>
 #include <gio/gunixinputstream.h>
@@ -44,6 +45,9 @@
 #include "resources.h"
 #include "synctex.h"
 #include "content-type.h"
+#ifdef WITH_SECCOMP
+#include "seccomp-filters.h"
+#endif
 
 typedef struct zathura_document_info_s {
   zathura_t* zathura;
@@ -86,6 +90,7 @@ zathura_create(void)
 
   /* global settings */
   zathura->global.search_direction = FORWARD;
+  zathura->global.sandbox = ZATHURA_SANDBOX_NORMAL;
 
   /* plugins */
   zathura->plugins.manager = zathura_plugin_manager_new();
@@ -99,6 +104,9 @@ zathura_create(void)
     goto error_out;
   }
 
+  /* set default icon */
+  girara_setting_set(zathura->ui.session, "window-icon", "org.pwmt.zathura");
+
 #ifdef G_OS_UNIX
   /* signal handler */
   zathura->signals.sigterm = g_unix_signal_add(SIGTERM, zathura_signal_sigterm, zathura);
@@ -134,10 +142,74 @@ create_directories(zathura_t* zathura)
   }
 }
 
+void
+zathura_update_view_ppi(zathura_t* zathura)
+{
+  if (zathura == NULL) {
+    return;
+  }
+
+  /* get view widget GdkMonitor */
+  GdkWindow* window = gtk_widget_get_window(zathura->ui.session->gtk.view); // NULL if not realized
+  if (window == NULL) {
+    return;
+  }
+  GdkDisplay* display = gtk_widget_get_display(zathura->ui.session->gtk.view);
+  if (display == NULL) {
+    return;
+  }
+
+  double ppi = 0.0;
+
+  GdkMonitor* monitor = gdk_display_get_monitor_at_window(display, window);
+  if (monitor == NULL) {
+    return;
+  }
+
+  /* physical width of monitor */
+  const int width_mm = gdk_monitor_get_width_mm(monitor);
+
+  /* size of monitor in pixels */
+  GdkRectangle monitor_geom;
+  gdk_monitor_get_geometry(monitor, &monitor_geom);
+
+  /* calculate ppi, knowing that 1 inch = 25.4 mm */
+  if (width_mm == 0) {
+    girara_debug("cannot calculate PPI: monitor has zero width");
+  } else {
+    ppi = monitor_geom.width * 25.4 / width_mm;
+  }
+
+#ifdef GDK_WINDOWING_WAYLAND
+  /* work around apparent bug in GDK: on Wayland, monitor geometry doesn't
+   * return values in application pixels as documented, but in device pixels.
+   * */
+  if (GDK_IS_WAYLAND_DISPLAY(display))
+  {
+    /* not using the cached value for the scale factor here to avoid issues
+     * if this function is called before the cached value is updated */
+    const int device_factor = gtk_widget_get_scale_factor(zathura->ui.session->gtk.view);
+    girara_debug("on Wayland, correcting PPI for device scale factor = %d", device_factor);
+    if (device_factor != 0) {
+      ppi /= device_factor;
+    }
+  }
+#endif
+
+  const double current_ppi = zathura_document_get_viewport_ppi(zathura->document);
+  if (fabs(ppi - current_ppi) > DBL_EPSILON) {
+    girara_debug("monitor width: %d mm, pixels: %d, ppi: %0.2f", width_mm, monitor_geom.width, ppi);
+    zathura_document_set_viewport_ppi(zathura->document, ppi);
+    render_all(zathura);
+    refresh_view(zathura);
+  }
+}
+
 static bool
 init_ui(zathura_t* zathura)
 {
   if (girara_session_init(zathura->ui.session, "zathura") == false) {
+    girara_error("Failed to initialize girara.");
     return false;
   }
 
@@ -153,11 +225,24 @@ init_ui(zathura_t* zathura)
   g_signal_connect(G_OBJECT(zathura->ui.session->gtk.view), "refresh-view",
                    G_CALLBACK(cb_refresh_view), zathura);
 
+  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.view),
+      "notify::scale-factor", G_CALLBACK(cb_scale_factor), zathura);
+
+  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.view),
+      "screen-changed", G_CALLBACK(cb_widget_screen_changed), zathura);
+
+  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window),
+      "configure-event", G_CALLBACK(cb_widget_configured), zathura);
+
+  /* initialize the screen-changed handler to 0 (i.e. invalid) */
+  zathura->signals.monitors_changed_handler = 0;
+
   /* page view */
   zathura->ui.page_widget = gtk_grid_new();
   gtk_grid_set_row_homogeneous(GTK_GRID(zathura->ui.page_widget), TRUE);
   gtk_grid_set_column_homogeneous(GTK_GRID(zathura->ui.page_widget), TRUE);
   if (zathura->ui.page_widget == NULL) {
+    girara_error("Failed to create page widget.");
     return false;
   }
 
@@ -197,18 +282,21 @@ init_ui(zathura_t* zathura)
   zathura->ui.statusbar.file =
     girara_statusbar_item_add(zathura->ui.session, TRUE, TRUE, TRUE, NULL);
   if (zathura->ui.statusbar.file == NULL) {
+    girara_error("Failed to create status bar item.");
     return false;
   }
 
   zathura->ui.statusbar.buffer =
     girara_statusbar_item_add(zathura->ui.session, FALSE, FALSE, FALSE, NULL);
   if (zathura->ui.statusbar.buffer == NULL) {
+    girara_error("Failed to create status bar item.");
     return false;
   }
 
   zathura->ui.statusbar.page_number =
     girara_statusbar_item_add(zathura->ui.session, FALSE, FALSE, FALSE, NULL);
   if (zathura->ui.statusbar.page_number == NULL) {
+    girara_error("Failed to create status bar item.");
     return false;
   }
 
@@ -222,13 +310,13 @@ init_ui(zathura_t* zathura)
   return true;
 }
 
-static void
+static bool
 init_css(zathura_t* zathura)
 {
   GiraraTemplate* csstemplate =
     girara_session_get_template(zathura->ui.session);
 
-  static const char* index_settings[] = {
+  static const char index_settings[][16] = {
     "index-fg",
     "index-bg",
     "index-active-fg",
@@ -242,7 +330,7 @@ init_css(zathura_t* zathura)
     GdkRGBA rgba      = {0, 0, 0, 0};
     girara_setting_get(zathura->ui.session, index_settings[s], &tmp_value);
     if (tmp_value != NULL) {
-      gdk_rgba_parse(&rgba, tmp_value);
+      parse_color(&rgba, tmp_value);
       g_free(tmp_value);
     }
 
@@ -261,7 +349,11 @@ init_css(zathura_t* zathura)
     girara_template_set_base(csstemplate, css);
     g_free(css);
     g_bytes_unref(css_data);
+  } else {
+    return false;
   }
+
+  return true;
 }
 
 static void
@@ -328,6 +420,9 @@ zathura_init(zathura_t* zathura)
     return false;
   }
 
+  /* Set application ID */
+  g_set_prgname("org.pwmt.zathura");
+
   /* create zathura (config/data) directory */
   create_directories(zathura);
 
@@ -338,8 +433,34 @@ zathura_init(zathura_t* zathura)
   config_load_default(zathura);
   config_load_files(zathura);
 
+#ifdef WITH_SECCOMP
+  /* initialize seccomp filters */
+  switch (zathura->global.sandbox) {
+    case ZATHURA_SANDBOX_NONE:
+      girara_debug("Sandbox deactivated.");
+      break;
+    case ZATHURA_SANDBOX_NORMAL:
+      girara_debug("Basic sandbox allowing normal operation.");
+      if (seccomp_enable_basic_filter() != 0) {
+        girara_error("Failed to initialize basic seccomp filter.");
+        goto error_free;
+      }
+      break;
+    case ZATHURA_SANDBOX_STRICT:
+      girara_debug("Strict sandbox preventing write and network access.");
+      if (seccomp_enable_strict_filter() != 0) {
+        girara_error("Failed to initialize strict seccomp filter.");
+        goto error_free;
+      }
+      /* unset the input method to avoid communication with external services */
+      unsetenv("GTK_IM_MODULE");
+      break;
+  }
+#endif
+
   /* UI */
-  if (!init_ui(zathura)) {
+  if (init_ui(zathura) == false) {
+    girara_error("Failed to initialize UI.");
     goto error_free;
   }
 
@@ -355,7 +476,10 @@ zathura_init(zathura_t* zathura)
   init_jumplist(zathura);
 
   /* CSS for index mode */
-  init_css(zathura);
+  if (init_css(zathura) == false) {
+    girara_error("Failed to initialize CSS.");
+    goto error_free;
+  }
 
   /* Shortcut helpers */
   init_shortcut_helpers(zathura);
@@ -434,14 +558,14 @@ zathura_free(zathura_t* zathura)
   g_free(zathura->config.cache_dir);
 
   /* free jumplist */
-  if (zathura->jumplist.list != NULL) {
-    girara_list_free(zathura->jumplist.list);
-  }
-
   if (zathura->jumplist.cur != NULL) {
     girara_list_iterator_free(zathura->jumplist.cur);
   }
 
+  if (zathura->jumplist.list != NULL) {
+    girara_list_free(zathura->jumplist.list);
+  }
+
   g_free(zathura);
 }
 
@@ -495,6 +619,23 @@ zathura_set_cache_dir(zathura_t* zathura, const char* dir)
   }
 }
 
+static void
+add_dir(void* data, void* userdata)
+{
+  const char* path                         = data;
+  zathura_plugin_manager_t* plugin_manager = userdata;
+
+  zathura_plugin_manager_add_dir(plugin_manager, path);
+}
+
+static void
+set_plugin_dir(zathura_t* zathura, const char* dir)
+{
+  girara_list_t* paths = girara_split_path_array(dir);
+  girara_list_foreach(paths, add_dir, zathura->plugins.manager);
+  girara_list_free(paths);
+}
+
 void
 zathura_set_plugin_dir(zathura_t* zathura, const char* dir)
 {
@@ -502,21 +643,14 @@ zathura_set_plugin_dir(zathura_t* zathura, const char* dir)
   g_return_if_fail(zathura->plugins.manager != NULL);
 
   if (dir != NULL) {
-    girara_list_t* paths = girara_split_path_array(dir);
-    GIRARA_LIST_FOREACH(paths, char*, iter, path)
-    zathura_plugin_manager_add_dir(zathura->plugins.manager, path);
-    GIRARA_LIST_FOREACH_END(paths, char*, iter, path);
-    girara_list_free(paths);
-  } else {
+    set_plugin_dir(zathura, dir);
+  } else if (g_getenv("ZATHURA_PLUGINS_PATH") != NULL) {
+    set_plugin_dir(zathura, g_getenv("ZATHURA_PLUGINS_PATH"));
 #ifdef ZATHURA_PLUGINDIR
-    girara_list_t* paths = girara_split_path_array(ZATHURA_PLUGINDIR);
-    GIRARA_LIST_FOREACH(paths, char*, iter, path)
-    zathura_plugin_manager_add_dir(zathura->plugins.manager, path);
-    GIRARA_LIST_FOREACH_END(paths, char*, iter, path);
-    girara_list_free(paths);
+  } else {
+    set_plugin_dir(zathura, ZATHURA_PLUGINDIR);
 #endif
   }
-
 }
 
 void
@@ -527,6 +661,58 @@ zathura_set_argv(zathura_t* zathura, char** argv)
   zathura->global.arguments = argv;
 }
 
+static bool
+setup_renderer(zathura_t* zathura, zathura_document_t* document)
+{
+ /* page cache size */
+  int cache_size = 0;
+  girara_setting_get(zathura->ui.session, "page-cache-size", &cache_size);
+  if (cache_size <= 0) {
+    girara_warning("page-cache-size is not positive, using %d instead",
+        ZATHURA_PAGE_CACHE_DEFAULT_SIZE);
+    cache_size = ZATHURA_PAGE_CACHE_DEFAULT_SIZE;
+  }
+
+  girara_debug("starting renderer with cache size %d", cache_size);
+  ZathuraRenderer* renderer = zathura_renderer_new(cache_size);
+  if (renderer == NULL) {
+    return false;
+  }
+
+  /* set up recolor info in ZathuraRenderer */
+  char* recolor_dark  = NULL;
+  char* recolor_light = NULL;
+  girara_setting_get(zathura->ui.session, "recolor-darkcolor", &recolor_dark);
+  girara_setting_get(zathura->ui.session, "recolor-lightcolor", &recolor_light);
+  zathura_renderer_set_recolor_colors_str(renderer, recolor_light, recolor_dark);
+  g_free(recolor_dark);
+  g_free(recolor_light);
+
+  bool recolor = false;
+  girara_setting_get(zathura->ui.session, "recolor", &recolor);
+  zathura_renderer_enable_recolor(renderer, recolor);
+  girara_setting_get(zathura->ui.session, "recolor-keephue", &recolor);
+  zathura_renderer_enable_recolor_hue(renderer, recolor);
+  girara_setting_get(zathura->ui.session, "recolor-reverse-video", &recolor);
+  zathura_renderer_enable_recolor_reverse_video(renderer, recolor);
+
+  zathura->sync.render_thread = renderer;
+
+  /* create render request to render window icon */
+  bool window_icon = false;
+  girara_setting_get(zathura->ui.session, "window-icon-document", &window_icon);
+  if (window_icon == true) {
+    girara_debug("starting render request for window icon");
+    ZathuraRenderRequest* request = zathura_render_request_new(renderer, zathura_document_get_page(document, 0));
+    g_signal_connect(request, "completed", G_CALLBACK(cb_window_update_icon), zathura);
+    zathura_render_request_set_render_plain(request, true);
+    zathura_render_request(request, 0);
+    zathura->window_icon_render_request = request;
+  }
+
+  return true;
+}
+
 #ifdef G_OS_UNIX
 static gchar*
 prepare_document_open_from_stdin(const char* path)
@@ -647,7 +833,11 @@ document_info_open(gpointer data)
         document_info->zathura->stdin_support.file = g_strdup(file);
       }
     } else {
-      GFile* gf = g_file_new_for_commandline_arg(document_info->path);
+      /* expand ~ and ~user in paths if present */
+      char* tmp_path = *document_info->path == '~' ? girara_fix_path(document_info->path) : NULL;
+      GFile* gf = g_file_new_for_commandline_arg(tmp_path != NULL ? tmp_path : document_info->path);
+      g_free(tmp_path);
+
       if (g_file_is_native(gf) == TRUE) {
         /* file was given as a native path */
         file = g_file_get_path(gf);
@@ -668,7 +858,7 @@ document_info_open(gpointer data)
 
     if (file != NULL) {
       if (document_info->synctex != NULL) {
-        document_open_synctex(document_info->zathura, file, uri, 
+        document_open_synctex(document_info->zathura, file, uri,
                               document_info->password, document_info->synctex);
       } else {
         document_open(document_info->zathura, file, uri, document_info->password,
@@ -760,8 +950,12 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     goto error_out;
   }
 
+  /* FIXME: since there are many call chains leading here, check again if we need to expand ~ or
+   * ~user. We should fix all call sites instead */
+  char* tmp_path = *path == '~' ? girara_fix_path(path) : NULL;
   zathura_error_t error = ZATHURA_ERROR_OK;
-  zathura_document_t* document = zathura_document_open(zathura, path, uri, password, &error);
+  zathura_document_t* document = zathura_document_open(zathura, tmp_path != NULL ? tmp_path : path, uri, password, &error);
+  g_free(tmp_path);
 
   if (document == NULL) {
     if (error == ZATHURA_ERROR_INVALID_PASSWORD) {
@@ -802,10 +996,11 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
   zathura_fileinfo_t file_info = {
     .current_page = 0,
     .page_offset = 0,
-    .scale = 1,
+    .zoom = 1,
     .rotation = 0,
     .pages_per_row = 0,
     .first_page_column_list = NULL,
+    .page_right_to_left = false,
     .position_x = 0,
     .position_y = 0
   };
@@ -817,12 +1012,12 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
   /* set page offset */
   zathura_document_set_page_offset(document, file_info.page_offset);
 
-  /* check for valid scale value */
-  if (file_info.scale <= DBL_EPSILON) {
-    file_info.scale = 1;
+  /* check for valid zoom value */
+  if (file_info.zoom <= DBL_EPSILON) {
+    file_info.zoom = 1;
   }
-  zathura_document_set_scale(document,
-      zathura_correct_scale_value(zathura->ui.session, file_info.scale));
+  zathura_document_set_zoom(document,
+      zathura_correct_zoom_value(zathura->ui.session, file_info.zoom));
 
   /* check current page number */
   /* if it wasn't specified on the command-line, get it from file_info */
@@ -882,8 +1077,11 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     char* filemonitor_backend = NULL;
     girara_setting_get(zathura->ui.session, "filemonitor", &filemonitor_backend);
     zathura_filemonitor_type_t type = ZATHURA_FILEMONITOR_GLIB;
+    if (g_strcmp0(filemonitor_backend, "noop") == 0) {
+      type = ZATHURA_FILEMONITOR_NOOP;
+    }
 #ifdef G_OS_UNIX
-    if (g_strcmp0(filemonitor_backend, "signal") == 0) {
+    else if (g_strcmp0(filemonitor_backend, "signal") == 0) {
       type = ZATHURA_FILEMONITOR_SIGNAL;
     }
 #endif
@@ -895,6 +1093,9 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     }
     g_signal_connect(G_OBJECT(zathura->file_monitor.monitor), "reload-file",
                      G_CALLBACK(cb_file_monitor), zathura->ui.session);
+
+    girara_debug("starting file monitor");
+    zathura_filemonitor_start(zathura->file_monitor.monitor);
   }
 
   if (password != NULL) {
@@ -911,40 +1112,11 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
     goto error_free;
   }
 
-  /* page cache size */
-  int cache_size = 0;
-  girara_setting_get(zathura->ui.session, "page-cache-size", &cache_size);
-  if (cache_size <= 0) {
-    girara_warning("page-cache-size is not positive, using %d instead",
-        ZATHURA_PAGE_CACHE_DEFAULT_SIZE);
-    cache_size = ZATHURA_PAGE_CACHE_DEFAULT_SIZE;
-  }
-
   /* threads */
-  zathura->sync.render_thread = zathura_renderer_new(cache_size);
-
-  if (zathura->sync.render_thread == NULL) {
+  if (!setup_renderer(zathura, document)) {
     goto error_free;
   }
 
-  /* set up recolor info in ZathuraRenderer */
-  char* recolor_dark  = NULL;
-  char* recolor_light = NULL;
-  girara_setting_get(zathura->ui.session, "recolor-darkcolor", &recolor_dark);
-  girara_setting_get(zathura->ui.session, "recolor-lightcolor", &recolor_light);
-  zathura_renderer_set_recolor_colors_str(zathura->sync.render_thread,
-      recolor_light, recolor_dark);
-  g_free(recolor_dark);
-  g_free(recolor_light);
-
-  bool recolor = false;
-  girara_setting_get(zathura->ui.session, "recolor", &recolor);
-  zathura_renderer_enable_recolor(zathura->sync.render_thread, recolor);
-  girara_setting_get(zathura->ui.session, "recolor-keephue", &recolor);
-  zathura_renderer_enable_recolor_hue(zathura->sync.render_thread, recolor);
-  girara_setting_get(zathura->ui.session, "recolor-reverse-video", &recolor);
-  zathura_renderer_enable_recolor_reverse_video(zathura->sync.render_thread, recolor);
-
   /* get view port size */
   GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(
                  GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
@@ -956,6 +1128,15 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
   const unsigned int view_height = (unsigned int)floor(gtk_adjustment_get_page_size(vadjustment));
   zathura_document_set_viewport_height(zathura->document, view_height);
 
+  zathura_update_view_ppi(zathura);
+
+  /* call screen-changed callback to connect monitors-changed signal on initial screen */
+  cb_widget_screen_changed(zathura->ui.session->gtk.view, NULL, zathura);
+
+  /* get initial device scale */
+  int device_factor = gtk_widget_get_scale_factor(zathura->ui.session->gtk.view);
+  zathura_document_set_device_factors(zathura->document, device_factor, device_factor);
+
   /* create blank pages */
   zathura->pages = calloc(number_of_pages, sizeof(GtkWidget*));
   if (zathura->pages == NULL) {
@@ -995,6 +1176,7 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
   unsigned int pages_per_row = 1;
   char* first_page_column_list = NULL;
   unsigned int page_padding = 1;
+  bool page_right_to_left = false;
 
   girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
 
@@ -1020,12 +1202,13 @@ document_open(zathura_t* zathura, const char* path, const char* uri, const char*
   g_free(file_info.first_page_column_list);
   g_free(first_page_column_list);
 
-  page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column);
+  page_right_to_left = file_info.page_right_to_left;
+
+  page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column, page_right_to_left);
   zathura_document_set_page_layout(zathura->document, page_padding, pages_per_row, first_page_column);
 
   girara_set_view(zathura->ui.session, zathura->ui.page_widget);
 
-
   /* bookmarks */
   if (zathura->database != NULL) {
     if (zathura_bookmarks_load(zathura, file_path) == false) {
@@ -1181,10 +1364,11 @@ save_fileinfo_to_db(zathura_t* zathura)
   zathura_fileinfo_t file_info = {
     .current_page = zathura_document_get_current_page_number(zathura->document),
     .page_offset = zathura_document_get_page_offset(zathura->document),
-    .scale = zathura_document_get_scale(zathura->document),
+    .zoom = zathura_document_get_zoom(zathura->document),
     .rotation = zathura_document_get_rotation(zathura->document),
     .pages_per_row = 1,
     .first_page_column_list = "1:2",
+    .page_right_to_left = false,
     .position_x = zathura_document_get_position_x(zathura->document),
     .position_y = zathura_document_get_position_y(zathura->document)
   };
@@ -1193,6 +1377,8 @@ save_fileinfo_to_db(zathura_t* zathura)
                      &(file_info.pages_per_row));
   girara_setting_get(zathura->ui.session, "first-page-column",
                      &(file_info.first_page_column_list));
+  girara_setting_get(zathura->ui.session, "page-right-to-left",
+                     &(file_info.page_right_to_left));
 
   /* save file info */
   zathura_db_set_fileinfo(zathura->database, path, &file_info);
@@ -1209,8 +1395,17 @@ document_close(zathura_t* zathura, bool keep_monitor)
     return false;
   }
 
+  /* reset window icon */
+  if (zathura->ui.session != NULL && zathura->window_icon_render_request != NULL) {
+    char* window_icon = NULL;
+    girara_setting_get(zathura->ui.session, "window-icon", &window_icon);
+    girara_setting_set(zathura->ui.session, "window-icon", window_icon);
+    g_free(window_icon);
+  }
+
   /* stop rendering */
   zathura_renderer_stop(zathura->sync.render_thread);
+  g_clear_object(&zathura->window_icon_render_request);
 
   /* remove monitor */
   if (keep_monitor == false) {
@@ -1233,6 +1428,7 @@ document_close(zathura_t* zathura, bool keep_monitor)
     save_fileinfo_to_db(zathura);
   }
 
+  /* remove jump list */
   girara_list_iterator_free(zathura->jumplist.cur);
   zathura->jumplist.cur = NULL;
   girara_list_free(zathura->jumplist.list);
@@ -1312,7 +1508,16 @@ statusbar_page_number_update(zathura_t* zathura)
   unsigned int current_page_number = zathura_document_get_current_page_number(zathura->document);
 
   if (zathura->document != NULL) {
-    char* page_number_text = g_strdup_printf("[%d/%d]", current_page_number + 1, number_of_pages);
+    zathura_page_t* page = zathura_document_get_page(zathura->document, current_page_number);
+    char* page_label = zathura_page_get_label(page, NULL);
+
+    char* page_number_text = NULL;
+    if (page_label != NULL) {
+      page_number_text = g_strdup_printf("[%s (%d/%d)]", page_label, current_page_number + 1, number_of_pages);
+      g_free(page_label);
+    } else {
+      page_number_text = g_strdup_printf("[%d/%d]", current_page_number + 1, number_of_pages);
+    }
     girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.page_number, page_number_text);
 
     bool page_number_in_window_title = false;
@@ -1334,7 +1539,8 @@ statusbar_page_number_update(zathura_t* zathura)
 
 void
 page_widget_set_mode(zathura_t* zathura, unsigned int page_padding,
-                     unsigned int pages_per_row, unsigned int first_page_column)
+                     unsigned int pages_per_row, unsigned int first_page_column,
+                     bool page_right_to_left)
 {
   /* show at least one page */
   if (pages_per_row == 0) {
@@ -1365,6 +1571,9 @@ page_widget_set_mode(zathura_t* zathura, unsigned int page_padding,
     int y = (i + first_page_column - 1) / pages_per_row;
 
     GtkWidget* page_widget = zathura->pages[i];
+    if (page_right_to_left) {
+      x = pages_per_row - 1 - x;
+    }
 
     gtk_grid_attach(GTK_GRID(zathura->ui.page_widget), page_widget, x, y, 1, 1);
   }
@@ -1387,11 +1596,11 @@ position_set(zathura_t* zathura, double position_x, double position_y)
 
   /* xalign = 0.5: center horizontally (with the page, not the document) */
   if (vertical_center == true) {
-    /* yalign = 0.0: align page an viewport edges at the top               */
-    page_number_to_position(zathura->document, page_id, 0.5, 0.0, &comppos_x, &comppos_y);
-  } else {
     /* yalign = 0.5: center vertically */
     page_number_to_position(zathura->document, page_id, 0.5, 0.5, &comppos_x, &comppos_y);
+  } else {
+    /* yalign = 0.0: align page an viewport edges at the top               */
+    page_number_to_position(zathura->document, page_id, 0.5, 0.0, &comppos_x, &comppos_y);
   }
 
   /* automatic horizontal adjustment */
@@ -1425,7 +1634,6 @@ position_set(zathura_t* zathura, double position_x, double position_y)
   return true;
 }
 
-
 void
 refresh_view(zathura_t* zathura)
 {
@@ -1436,7 +1644,6 @@ refresh_view(zathura_t* zathura)
                 0, zathura);
 }
 
-
 bool
 adjust_view(zathura_t* zathura)
 {
@@ -1466,33 +1673,31 @@ adjust_view(zathura_t* zathura)
 
   double page_ratio = (double)cell_height / (double)document_width;
   double view_ratio = (double)view_height / (double)view_width;
-  double scale = zathura_document_get_scale(zathura->document);
-  double newscale = scale;
+  double zoom = zathura_document_get_zoom(zathura->document);
+  double newzoom = zoom;
 
   if (adjust_mode == ZATHURA_ADJUST_WIDTH ||
       (adjust_mode == ZATHURA_ADJUST_BESTFIT && page_ratio < view_ratio)) {
-    newscale *= (double)view_width / (double)document_width;
-
+    newzoom *= (double)view_width / (double)document_width;
   } else if (adjust_mode == ZATHURA_ADJUST_BESTFIT) {
-    newscale *= (double)view_height / (double)cell_height;
-
+    newzoom *= (double)view_height / (double)cell_height;
   } else {
     goto error_ret;
   }
 
-  /* save new scale and recompute cell size */
-  zathura_document_set_scale(zathura->document, newscale);
+  /* save new zoom and recompute cell size */
+  zathura_document_set_zoom(zathura->document, newzoom);
   unsigned int new_cell_height = 0, new_cell_width = 0;
   zathura_document_get_cell_size(zathura->document, &new_cell_height, &new_cell_width);
 
-  /* if the change in scale changes page cell dimensions by at least one pixel, render */
+  /* if the change in zoom changes page cell dimensions by at least one pixel, render */
   if (abs((int)new_cell_width - (int)cell_width) > 1 ||
       abs((int)new_cell_height - (int)cell_height) > 1) {
     render_all(zathura);
     refresh_view(zathura);
   } else {
-    /* otherwise set the old scale and leave */
-    zathura_document_set_scale(zathura->document, scale);
+    /* otherwise set the old zoom and leave */
+    zathura_document_set_zoom(zathura->document, zoom);
   }
 
 error_ret:
@@ -1503,13 +1708,11 @@ error_ret:
 static gboolean
 zathura_signal_sigterm(gpointer data)
 {
-  if (data == NULL) {
-    return TRUE;
+  if (data != NULL) {
+    zathura_t* zathura = data;
+    cb_destroy(NULL, zathura);
   }
 
-  zathura_t* zathura = (zathura_t*) data;
-  cb_destroy(NULL, zathura);
-
   return TRUE;
 }
 #endif
